#!/usr/bin/python

from __future__ import annotations

import gc
import unittest
from datetime import date, datetime, time, timedelta, timezone
from decimal import Decimal
from typing import Any, ClassVar
from uuid import UUID as Uuid  # noqa: N811

import pgdb

from . import dbapi20
from .config import dbhost, dbname, dbpasswd, dbport, dbuser


class PgBitString:
    """Test object with a PostgreSQL representation as Bit String."""

    def __init__(self, value):
        self.value = value

    def __pg_repr__(self):
        return f"B'{self.value:b}'"


class TestPgDb(dbapi20.DatabaseAPI20Test):

    driver = pgdb
    connect_args = ()
    connect_kw_args: ClassVar[dict[str, Any]] = {
        'database': dbname, 'host': f"{dbhost or ''}:{dbport or -1}",
        'user': dbuser, 'password': dbpasswd}

    lower_func = 'lower'  # For stored procedure test

    def setUp(self):
        super().setUp()
        try:
            con = self._connect()
            con.close()
        except pgdb.Error:  # try to create a missing database
            import pg
            try:  # first try to log in as superuser
                db = pg.DB('postgres', dbhost or None, dbport or -1,
                           user='postgres')
            except Exception:  # then try to log in as current user
                db = pg.DB('postgres', dbhost or None, dbport or -1)
            db.query('create database ' + dbname)

    def tearDown(self):
        super().tearDown()

    def test_version(self):
        v = pgdb.version
        self.assertIsInstance(v, str)
        self.assertIn('.', v)
        self.assertEqual(pgdb.__version__, v)

    def test_connect_kwargs(self):
        application_name = 'PyGreSQL DB API 2.0 Test'
        self.connect_kw_args['application_name'] = application_name
        con = self._connect()
        cur = con.cursor()
        cur.execute("select application_name from pg_stat_activity"
                    " where application_name = %s", (application_name,))
        self.assertEqual(cur.fetchone(), (application_name,))

    def test_connect_kwargs_with_special_chars(self):
        special_name = 'Single \' and double " quote and \\ backslash!'
        self.connect_kw_args['application_name'] = special_name
        con = self._connect()
        cur = con.cursor()
        cur.execute("select application_name from pg_stat_activity"
                    " where application_name = %s", (special_name,))
        self.assertEqual(cur.fetchone(), (special_name,))

    def test_percent_sign(self):
        con = self._connect()
        cur = con.cursor()
        cur.execute("select %s, 'a %% sign'", ('a % sign',))
        self.assertEqual(cur.fetchone(), ('a % sign', 'a % sign'))
        cur.execute("select 'a % sign'")
        self.assertEqual(cur.fetchone(), ('a % sign',))
        cur.execute("select 'a %% sign'")
        self.assertEqual(cur.fetchone(), ('a % sign',))

    def test_paramstyles(self):
        self.assertEqual(pgdb.paramstyle, 'pyformat')
        con = self._connect()
        cur = con.cursor()
        # parameters can be passed as tuple
        cur.execute("select %s, %s, %s", (123, 'abc', True))
        self.assertEqual(cur.fetchone(), (123, 'abc', True))
        # parameters can be passed as dict
        cur.execute("select %(one)s, %(two)s, %(one)s, %(three)s", {
            "one": 123, "two": "abc", "three": True
        })
        self.assertEqual(cur.fetchone(), (123, 'abc', 123, True))

    def test_callproc_no_params(self):
        con = self._connect()
        cur = con.cursor()
        # note that now() does not change within a transaction
        cur.execute('select now()')
        now = cur.fetchone()[0]
        res = cur.callproc('now')
        self.assertIsNone(res)
        res = cur.fetchone()[0]
        self.assertEqual(res, now)

    def test_callproc_bad_params(self):
        con = self._connect()
        cur = con.cursor()
        self.assertRaises(TypeError, cur.callproc, 'lower', 42)
        self.assertRaises(pgdb.ProgrammingError, cur.callproc, 'lower', (42,))

    def test_callproc_one_param(self):
        con = self._connect()
        cur = con.cursor()
        params = (42.4382,)
        res = cur.callproc("round", params)
        self.assertIs(res, params)
        res = cur.fetchone()[0]
        self.assertEqual(res, 42)

    def test_callproc_two_params(self):
        con = self._connect()
        cur = con.cursor()
        params = (9, 4)
        res = cur.callproc("div", params)
        self.assertIs(res, params)
        res = cur.fetchone()[0]
        self.assertEqual(res, 2)

    def test_cursor_type(self):

        class TestCursor(pgdb.Cursor):
            @staticmethod
            def row_factory(row):
                return row  # not used

        con = self._connect()
        self.assertIs(con.cursor_type, pgdb.Cursor)
        cur = con.cursor()
        self.assertIsInstance(cur, pgdb.Cursor)
        self.assertNotIsInstance(cur, TestCursor)
        con.cursor_type = TestCursor
        cur = con.cursor()
        self.assertIsInstance(cur, TestCursor)
        cur = con.cursor()
        self.assertIsInstance(cur, TestCursor)
        con = self._connect()
        self.assertIs(con.cursor_type, pgdb.Cursor)
        cur = con.cursor()
        self.assertIsInstance(cur, pgdb.Cursor)
        self.assertNotIsInstance(cur, TestCursor)

    def test_row_factory(self):

        class TestCursor(pgdb.Cursor):

            def row_factory(self, row):  # type: ignore[override]
                description = self.description
                assert isinstance(description, list)
                return {f'column {desc[0]}': value
                        for desc, value in zip(description, row)}

        con = self._connect()
        con.cursor_type = TestCursor
        cur = con.cursor()
        self.assertIsInstance(cur, TestCursor)
        res = cur.execute("select 1 as a, 2 as b")
        self.assertIs(res, cur, 'execute() should return cursor')
        res = cur.fetchone()
        self.assertIsInstance(res, dict)
        self.assertEqual(res, {'column a': 1, 'column b': 2})
        cur.execute("select 1 as a, 2 as b union select 3, 4 order by 1")
        res = cur.fetchall()
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], dict)
        self.assertEqual(res[0], {'column a': 1, 'column b': 2})
        self.assertIsInstance(res[1], dict)
        self.assertEqual(res[1], {'column a': 3, 'column b': 4})

    def test_build_row_factory(self):

        # noinspection PyAbstractClass
        class TestCursor(pgdb.Cursor):

            def build_row_factory(self):
                description = self.description
                assert isinstance(description, list)
                keys = [desc[0] for desc in description]
                return lambda row: {
                    key: value for key, value in zip(keys, row)}

        con = self._connect()
        con.cursor_type = TestCursor
        cur = con.cursor()
        self.assertIsInstance(cur, TestCursor)
        cur.execute("select 1 as a, 2 as b")
        res = cur.fetchone()
        self.assertIsInstance(res, dict)
        self.assertEqual(res, {'a': 1, 'b': 2})
        cur.execute("select 1 as a, 2 as b union select 3, 4 order by 1")
        res = cur.fetchall()
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], dict)
        self.assertEqual(res[0], {'a': 1, 'b': 2})
        self.assertIsInstance(res[1], dict)
        self.assertEqual(res[1], {'a': 3, 'b': 4})

    # noinspection PyUnresolvedReferences
    def test_cursor_with_named_columns(self):
        con = self._connect()
        cur = con.cursor()
        res = cur.execute("select 1 as abc, 2 as de, 3 as f")
        self.assertIs(res, cur, 'execute() should return cursor')
        res = cur.fetchone()
        self.assertIsInstance(res, tuple)
        self.assertEqual(res, (1, 2, 3))
        self.assertEqual(res._fields, ('abc', 'de', 'f'))
        self.assertEqual(res.abc, 1)
        self.assertEqual(res.de, 2)
        self.assertEqual(res.f, 3)
        cur.execute("select 1 as one, 2 as two union select 3, 4 order by 1")
        res = cur.fetchall()
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], tuple)
        self.assertEqual(res[0], (1, 2))
        self.assertEqual(res[0]._fields, ('one', 'two'))
        self.assertIsInstance(res[1], tuple)
        self.assertEqual(res[1], (3, 4))
        self.assertEqual(res[1]._fields, ('one', 'two'))

    # noinspection PyUnresolvedReferences
    def test_cursor_with_unnamed_columns(self):
        con = self._connect()
        cur = con.cursor()
        cur.execute("select 1, 2, 3")
        res = cur.fetchone()
        self.assertIsInstance(res, tuple)
        self.assertEqual(res, (1, 2, 3))
        self.assertEqual(res._fields, ('_0', '_1', '_2'))
        cur.execute("select 1 as one, 2, 3 as three")
        res = cur.fetchone()
        self.assertIsInstance(res, tuple)
        self.assertEqual(res, (1, 2, 3))
        self.assertEqual(res._fields, ('one', '_1', 'three'))

    # noinspection PyUnresolvedReferences
    def test_cursor_with_badly_named_columns(self):
        con = self._connect()
        cur = con.cursor()
        cur.execute("select 1 as abc, 2 as def")
        res = cur.fetchone()
        self.assertIsInstance(res, tuple)
        self.assertEqual(res, (1, 2))
        self.assertEqual(res._fields, ('abc', '_1'))
        cur.execute(
            'select 1 as snake_case, 2 as "CamelCase",'
            ' 3 as "kebap-case", 4 as "_bad", 5 as "0bad", 6 as "bad$"')
        res = cur.fetchone()
        self.assertIsInstance(res, tuple)
        self.assertEqual(res, (1, 2, 3, 4, 5, 6))
        self.assertEqual(res._fields[:2], ('snake_case', 'CamelCase'))
        fields = ('_2', '_3', '_4', '_5')
        self.assertEqual(res._fields[2:], fields)

    def test_colnames(self):
        con = self._connect()
        cur = con.cursor()
        cur.execute("select 1, 2, 3")
        names = cur.colnames
        self.assertIsInstance(names, list)
        self.assertEqual(names, ['?column?', '?column?', '?column?'])
        cur.execute("select 1 as a, 2 as bc, 3 as def, 4 as g")
        names = cur.colnames
        self.assertIsInstance(names, list)
        self.assertEqual(names, ['a', 'bc', 'def', 'g'])

    def test_coltypes(self):
        con = self._connect()
        cur = con.cursor()
        cur.execute("select 1::int2, 2::int4, 3::int8")
        types = cur.coltypes
        self.assertIsInstance(types, list)
        self.assertEqual(types, ['int2', 'int4', 'int8'])

    # noinspection PyUnresolvedReferences
    def test_description_fields(self):
        con = self._connect()
        cur = con.cursor()
        cur.execute("select 123456789::int8 col0,"
                    " 123456.789::numeric(41, 13) as col1,"
                    " 'foobar'::char(39) as col2")
        desc = cur.description
        self.assertIsInstance(desc, list)
        self.assertEqual(len(desc), 3)
        cols = [('int8', 8, None), ('numeric', 41, 13), ('bpchar', 39, None)]
        for i in range(3):
            c, d = cols[i], desc[i]
            self.assertIsInstance(d, tuple)
            self.assertEqual(len(d), 7)
            self.assertIsInstance(d.name, str)
            self.assertEqual(d.name, f'col{i}')
            self.assertIsInstance(d.type_code, str)
            self.assertEqual(d.type_code, c[0])
            self.assertIsNone(d.display_size)
            self.assertIsInstance(d.internal_size, int)
            self.assertEqual(d.internal_size, c[1])
            if c[2] is not None:
                self.assertIsInstance(d.precision, int)
                self.assertEqual(d.precision, c[1])
                self.assertIsInstance(d.scale, int)
                self.assertEqual(d.scale, c[2])
            else:
                self.assertIsNone(d.precision)
                self.assertIsNone(d.scale)
            self.assertIsNone(d.null_ok)

    def test_type_cache_info(self):
        con = self._connect()
        try:
            cur = con.cursor()
            type_cache = con.type_cache
            self.assertNotIn('numeric', type_cache)
            type_info = type_cache['numeric']
            self.assertIn('numeric', type_cache)
            self.assertEqual(type_info, 'numeric')
            self.assertEqual(type_info.oid, 1700)
            self.assertEqual(type_info.len, -1)
            self.assertEqual(type_info.type, 'b')  # base
            self.assertEqual(type_info.category, 'N')  # numeric
            self.assertEqual(type_info.delim, ',')
            self.assertEqual(type_info.relid, 0)
            self.assertIs(con.type_cache[1700], type_info)
            self.assertNotIn('pg_type', type_cache)
            type_info = type_cache['pg_type']
            self.assertIn('pg_type', type_cache)
            self.assertEqual(type_info.type, 'c')  # composite
            self.assertEqual(type_info.category, 'C')  # composite
            cols = type_cache.get_fields('pg_type')
            if cols[0].name == 'oid':  # PostgreSQL < 12
                del cols[0]
            self.assertEqual(cols[0].name, 'typname')
            typname = type_cache[cols[0].type]
            self.assertEqual(typname, 'name')
            self.assertEqual(typname.type, 'b')  # base
            self.assertEqual(typname.category, 'S')  # string
            self.assertEqual(cols[3].name, 'typlen')
            typlen = type_cache[cols[3].type]
            self.assertEqual(typlen, 'int2')
            self.assertEqual(typlen.type, 'b')  # base
            self.assertEqual(typlen.category, 'N')  # numeric
            cur.close()
            cur = con.cursor()
            type_cache = con.type_cache
            self.assertIn('numeric', type_cache)
            cur.close()
        finally:
            con.close()
        con = self._connect()
        try:
            cur = con.cursor()
            type_cache = con.type_cache
            self.assertNotIn('pg_type', type_cache)
            self.assertEqual(type_cache.get('pg_type'), type_info)
            self.assertIn('pg_type', type_cache)
            self.assertIsNone(type_cache.get(
                self.table_prefix + '_surely_does_not_exist'))
            cur.close()
        finally:
            con.close()

    def test_type_cache_typecast(self):
        con = self._connect()
        try:
            cur = con.cursor()
            type_cache = con.type_cache
            self.assertIs(type_cache.get_typecast('int4'), int)
            cast_int = lambda v: f'int({v})'  # noqa: E731
            type_cache.set_typecast('int4', cast_int)
            query = 'select 2::int2, 4::int4, 8::int8'
            cur.execute(query)
            i2, i4, i8 = cur.fetchone()
            self.assertEqual(i2, 2)
            self.assertEqual(i4, 'int(4)')
            self.assertEqual(i8, 8)
            self.assertEqual(type_cache.typecast(42, 'int4'), 'int(42)')
            type_cache.set_typecast(['int2', 'int8'], cast_int)
            cur.execute(query)
            i2, i4, i8 = cur.fetchone()
            self.assertEqual(i2, 'int(2)')
            self.assertEqual(i4, 'int(4)')
            self.assertEqual(i8, 'int(8)')
            type_cache.reset_typecast('int4')
            cur.execute(query)
            i2, i4, i8 = cur.fetchone()
            self.assertEqual(i2, 'int(2)')
            self.assertEqual(i4, 4)
            self.assertEqual(i8, 'int(8)')
            type_cache.reset_typecast(['int2', 'int8'])
            cur.execute(query)
            i2, i4, i8 = cur.fetchone()
            self.assertEqual(i2, 2)
            self.assertEqual(i4, 4)
            self.assertEqual(i8, 8)
            type_cache.set_typecast(['int2', 'int8'], cast_int)
            cur.execute(query)
            i2, i4, i8 = cur.fetchone()
            self.assertEqual(i2, 'int(2)')
            self.assertEqual(i4, 4)
            self.assertEqual(i8, 'int(8)')
            type_cache.reset_typecast()
            cur.execute(query)
            i2, i4, i8 = cur.fetchone()
            self.assertEqual(i2, 2)
            self.assertEqual(i4, 4)
            self.assertEqual(i8, 8)
            cur.close()
        finally:
            con.close()

    def test_cursor_iteration(self):
        con = self._connect()
        cur = con.cursor()
        cur.execute("select 1 union select 2 union select 3 order by 1")
        self.assertEqual([r[0] for r in cur], [1, 2, 3])

    def test_cursor_invalidation(self):
        con = self._connect()
        cur = con.cursor()
        cur.execute("select 1 union select 2")
        self.assertEqual(cur.fetchone(), (1,))
        self.assertFalse(con.closed)
        con.close()
        self.assertTrue(con.closed)
        self.assertRaises(pgdb.OperationalError, cur.fetchone)

    def test_fetch_2_rows(self):
        values = ('test', pgdb.Binary(b'\xff\x52\xb2'),
                  True, 5, 6, 5.7, Decimal('234.234234'), Decimal('75.45'),
                  pgdb.Date(2011, 7, 17), pgdb.Time(15, 47, 42),
                  pgdb.Timestamp(2008, 10, 20, 15, 25, 35),
                  pgdb.Interval(15, 31, 5), 7897234)
        table = self.table_prefix + 'booze'
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute("set datestyle to iso")
            cur.execute(
                f"create table {table} ("
                "stringtest varchar,"
                "binarytest bytea,"
                "booltest bool,"
                "integertest int4,"
                "longtest int8,"
                "floattest float8,"
                "numerictest numeric,"
                "moneytest money,"
                "datetest date,"
                "timetest time,"
                "datetimetest timestamp,"
                "intervaltest interval,"
                "rowidtest oid)")
            cur.execute("set standard_conforming_strings to on")
            for s in ('numeric', 'monetary', 'time'):
                cur.execute(f"set lc_{s} to 'C'")
            for _i in range(2):
                cur.execute(
                    f"insert into {table} values ("
                    "%s,%s,%s,%s,%s,%s,%s,"
                    "'%s'::money,%s,%s,%s,%s,%s)", values)
            cur.execute(f"select * from {table}")
            rows = cur.fetchall()
            self.assertEqual(len(rows), 2)
            row0 = rows[0]
            self.assertEqual(row0, values)
            self.assertEqual(row0, rows[1])
            self.assertIsInstance(row0[0], str)
            self.assertIsInstance(row0[1], bytes)
            self.assertIsInstance(row0[2], bool)
            self.assertIsInstance(row0[3], int)
            self.assertIsInstance(row0[4], int)
            self.assertIsInstance(row0[5], float)
            self.assertIsInstance(row0[6], Decimal)
            self.assertIsInstance(row0[7], Decimal)
            self.assertIsInstance(row0[8], date)
            self.assertIsInstance(row0[9], time)
            self.assertIsInstance(row0[10], datetime)
            self.assertIsInstance(row0[11], timedelta)
        finally:
            con.close()

    def test_integrity_error(self):
        table = self.table_prefix + 'booze'
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute("set client_min_messages = warning")
            cur.execute(f"create table {table} (i int primary key)")
            cur.execute(f"insert into {table} values (1)")
            cur.execute(f"insert into {table} values (2)")
            self.assertRaises(
                pgdb.IntegrityError, cur.execute,
                f"insert into {table} values (1)")
        finally:
            con.close()

    def test_update_rowcount(self):
        table = self.table_prefix + 'booze'
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute(f"create table {table} (i int)")
            cur.execute(f"insert into {table} values (1)")
            cur.execute(f"update {table} set i=2 where i=2 returning i")
            self.assertEqual(cur.rowcount, 0)
            cur.execute(f"update {table} set i=2 where i=1 returning i")
            self.assertEqual(cur.rowcount, 1)
            cur.close()
            # keep rowcount even if cursor is closed (needed by SQLAlchemy)
            self.assertEqual(cur.rowcount, 1)
        finally:
            con.close()

    def test_sqlstate(self):
        con = self._connect()
        cur = con.cursor()
        try:
            cur.execute("select 1/0")
        except pgdb.DatabaseError as error:
            self.assertIsInstance(error, pgdb.DataError)
            # the SQLSTATE error code for division by zero is 22012
            # noinspection PyUnresolvedReferences
            self.assertEqual(error.sqlstate, '22012')

    def test_float(self):
        nan, inf = float('nan'), float('inf')
        from math import isinf, isnan
        self.assertTrue(isnan(nan) and not isinf(nan))
        self.assertTrue(isinf(inf) and not isnan(inf))
        values = [0, 1, 0.03125, -42.53125, nan, inf, -inf,
                  'nan', 'inf', '-inf', 'NaN', 'Infinity', '-Infinity']
        table = self.table_prefix + 'booze'
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute(
                f"create table {table} (n smallint, floattest float)")
            params = enumerate(values)
            cur.executemany(f"insert into {table} values (%d,%s)", params)
            cur.execute(f"select floattest from {table} order by n")
            rows = cur.fetchall()
            self.assertEqual(cur.description[0].type_code, pgdb.FLOAT)
            self.assertNotEqual(cur.description[0].type_code, pgdb.ARRAY)
            self.assertNotEqual(cur.description[0].type_code, pgdb.RECORD)
        finally:
            con.close()
        self.assertEqual(len(rows), len(values))
        rows = [row[0] for row in rows]
        for inval, outval in zip(values, rows):
            if inval in ('inf', 'Infinity'):
                inval = inf
            elif inval in ('-inf', '-Infinity'):
                inval = -inf
            elif inval in ('nan', 'NaN'):
                inval = nan
            if isinf(inval):  # type: ignore
                self.assertTrue(isinf(outval))
                if inval < 0:  # type: ignore
                    self.assertLess(outval, 0)
                else:
                    self.assertGreater(outval, 0)
            elif isnan(inval):  # type: ignore
                self.assertTrue(isnan(outval))
            else:
                self.assertEqual(inval, outval)

    def test_datetime(self):
        dt = datetime(2011, 7, 17, 15, 47, 42, 317509)
        values = [dt.date(), dt.time(), dt, dt.time(), dt]
        self.assertIsInstance(values[3], time)
        assert isinstance(values[3], time)  # type guard
        values[3] = values[3].replace(tzinfo=timezone.utc)
        self.assertIsInstance(values[4], datetime)
        assert isinstance(values[4], datetime)  # type guard
        values[4] = values[4].replace(tzinfo=timezone.utc)
        da = (dt.year, dt.month, dt.day)
        ti = (dt.hour, dt.minute, dt.second, dt.microsecond)
        tz = (timezone.utc,)
        inputs = [
            # input as objects
            values,
            # input as text
            [v.isoformat() for v in values],  # type: ignore
            # # input using type helpers
            [pgdb.Date(*da), pgdb.Time(*ti),
             pgdb.Timestamp(*(da + ti)), pgdb.Time(*(ti + tz)),
             pgdb.Timestamp(*(da + ti + tz))]
        ]
        table = self.table_prefix + 'booze'
        con: pgdb.Connection = self._connect()
        try:
            cur = con.cursor()
            cur.execute("set timezone = UTC")
            cur.execute(f"create table {table} ("
                        "d date, t time,  ts timestamp,"
                        "tz timetz, tsz timestamptz)")
            for params in inputs:
                for datestyle in ('iso', 'postgres, mdy', 'postgres, dmy',
                                  'sql, mdy', 'sql, dmy', 'german'):
                    cur.execute(f"set datestyle to {datestyle}")
                    if not isinstance(params[0], str):
                        cur.execute("select %s,%s,%s,%s,%s", params)
                        row = cur.fetchone()
                        self.assertEqual(row, tuple(values))
                    cur.execute(
                        f"insert into {table}"
                        " values (%s,%s,%s,%s,%s)", params)
                    cur.execute(f"select * from {table}")
                    d = cur.description
                    self.assertIsInstance(d, list)
                    assert d is not None  # type guard
                    for i in range(5):
                        tc = d[i].type_code
                        self.assertEqual(tc, pgdb.DATETIME)
                        self.assertNotEqual(tc, pgdb.STRING)
                        self.assertNotEqual(tc, pgdb.ARRAY)
                        self.assertNotEqual(tc, pgdb.RECORD)
                    self.assertEqual(d[0].type_code, pgdb.DATE)
                    self.assertEqual(d[1].type_code, pgdb.TIME)
                    self.assertEqual(d[2].type_code, pgdb.TIMESTAMP)
                    self.assertEqual(d[3].type_code, pgdb.TIME)
                    self.assertEqual(d[4].type_code, pgdb.TIMESTAMP)
                    row = cur.fetchone()
                    self.assertEqual(row, tuple(values))
                    cur.execute(f"truncate table {table}")
        finally:
            con.close()

    def test_interval(self):
        td = datetime(2011, 7, 17, 15, 47, 42, 317509) - datetime(1970, 1, 1)
        inputs = [
            # input as objects
            td,
            # input as text
            f'{td.days} days {td.seconds} seconds'
            f' {td.microseconds} microseconds',
            # input using type helpers
            pgdb.Interval(td.days, 0, 0, td.seconds, td.microseconds)]
        table = self.table_prefix + 'booze'
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute(f"create table {table} (i interval)")
            for param in inputs:
                for intervalstyle in ('sql_standard ', 'postgres',
                                      'postgres_verbose', 'iso_8601'):
                    cur.execute(f"set intervalstyle to {intervalstyle}")
                    # noinspection PyUnboundLocalVariable
                    cur.execute(f"insert into {table} values (%s)", [param])
                    cur.execute(f"select * from {table}")
                    tc = cur.description[0].type_code
                    self.assertEqual(tc, pgdb.DATETIME)
                    self.assertNotEqual(tc, pgdb.STRING)
                    self.assertNotEqual(tc, pgdb.ARRAY)
                    self.assertNotEqual(tc, pgdb.RECORD)
                    self.assertEqual(tc, pgdb.INTERVAL)
                    row = cur.fetchone()
                    self.assertEqual(row, (td,))
                    cur.execute(f"truncate table {table}")
        finally:
            con.close()

    def test_hstore(self):
        con = self._connect()
        cur = con.cursor()
        try:
            cur.execute("select 'k=>v'::hstore")
        except pgdb.DatabaseError:
            try:
                cur.execute("create extension hstore")
            except pgdb.DatabaseError:
                self.skipTest("hstore extension not enabled")
        finally:
            con.close()
        d = {'k': 'v', 'foo': 'bar', 'baz': 'whatever', 'back\\': '\\slash',
             '1a': 'anything at all', '2=b': 'value = 2', '3>c': 'value > 3',
             '4"c': 'value " 4', "5'c": "value ' 5", 'hello, world': '"hi!"',
             'None': None, 'NULL': 'NULL', 'empty': ''}
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute("select %s::hstore", (pgdb.Hstore(d),))
            result = cur.fetchone()[0]
        finally:
            con.close()
        self.assertIsInstance(result, dict)
        self.assertEqual(result, d)

    def test_uuid(self):
        self.assertIs(Uuid, pgdb.Uuid)
        d = Uuid('{12345678-1234-5678-1234-567812345678}')
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute("select %s::uuid", (d,))
            result = cur.fetchone()[0]
        finally:
            con.close()
        self.assertIsInstance(result, Uuid)
        self.assertEqual(result, d)

    def test_insert_array(self):
        values: list[tuple[Any, Any]] = [
            (None, None), ([], []), ([None], [[None], ['null']]),
            ([1, 2, 3], [['a', 'b'], ['c', 'd']]),
            ([20000, 25000, 25000, 30000],
             [['breakfast', 'consulting'], ['meeting', 'lunch']]),
            ([0, 1, -1], [['Hello, World!', '"Hi!"'], ['{x,y}', ' x y ']])]
        table = self.table_prefix + 'booze'
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute(
                f"create table {table} (n smallint, i int[], t text[][])")
            params = [(n, v[0], v[1]) for n, v in enumerate(values)]
            # Note that we must explicit casts because we are inserting
            # empty arrays.  Otherwise this is not necessary.
            cur.executemany(
                f"insert into {table} values"
                " (%d,%s::int[],%s::text[][])", params)
            cur.execute(f"select i, t from {table} order by n")
            d = cur.description
            self.assertEqual(d[0].type_code, pgdb.ARRAY)
            self.assertNotEqual(d[0].type_code, pgdb.RECORD)
            self.assertEqual(d[0].type_code, pgdb.NUMBER)
            self.assertEqual(d[0].type_code, pgdb.INTEGER)
            self.assertEqual(d[1].type_code, pgdb.ARRAY)
            self.assertNotEqual(d[1].type_code, pgdb.RECORD)
            self.assertEqual(d[1].type_code, pgdb.STRING)
            rows = cur.fetchall()
        finally:
            con.close()
        self.assertEqual(rows, values)

    def test_select_array(self):
        values = ([1, 2, 3, None], ['a', 'b', 'c', None])
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute("select %s::int[], %s::text[]", values)
            row = cur.fetchone()
        finally:
            con.close()
        self.assertEqual(row, values)

    def test_unicode_list_and_tuple(self):
        value = ('Käse', 'Würstchen')
        con = self._connect()
        try:
            cur = con.cursor()
            try:
                cur.execute("select %s, %s", value)
            except pgdb.DatabaseError:
                self.skipTest('database does not support latin-1')
            row = cur.fetchone()
            cur.execute("select %s, %s", (list(value), tuple(value)))
            as_list, as_tuple = cur.fetchone()
        finally:
            con.close()
        self.assertEqual(as_list, list(row))
        self.assertEqual(as_tuple, tuple(row))

    def test_insert_record(self):
        values = [('John', 61), ('Jane', 63),
                  ('Fred', None), ('Wilma', None),
                  (None, 42), (None, None)]
        table = self.table_prefix + 'booze'
        record = self.table_prefix + 'munch'
        con = self._connect()
        cur = con.cursor()
        try:
            cur.execute(f"create type {record} as (name varchar, age int)")
            cur.execute(f"create table {table} (n smallint, r {record})")
            params = enumerate(values)
            cur.executemany(f"insert into {table} values (%d,%s)", params)
            cur.execute(f"select r from {table} order by n")
            type_code = cur.description[0].type_code
            self.assertEqual(type_code, record)
            self.assertEqual(type_code, pgdb.RECORD)
            self.assertNotEqual(type_code, pgdb.ARRAY)
            columns = con.type_cache.get_fields(type_code)
            self.assertEqual(columns[0].name, 'name')
            self.assertEqual(columns[1].name, 'age')
            self.assertEqual(con.type_cache[columns[0].type], 'varchar')
            self.assertEqual(con.type_cache[columns[1].type], 'int4')
            rows = cur.fetchall()
        finally:
            cur.execute(f'drop table {table}')
            cur.execute(f'drop type {record}')
            con.close()
        self.assertEqual(len(rows), len(values))
        rows = [row[0] for row in rows]
        self.assertEqual(rows, values)
        self.assertEqual(rows[0].name, 'John')
        self.assertEqual(rows[0].age, 61)

    def test_select_record(self):
        value = (1, 25000, 2.5, 'hello', 'Hello World!', 'Hello, World!',
                 '(test)', '(x,y)', ' x y ', 'null', None)
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute("select %s as test_record", [value])
            self.assertEqual(cur.description[0].name, 'test_record')
            self.assertEqual(cur.description[0].type_code, 'record')
            row = cur.fetchone()[0]
        finally:
            con.close()
        # Note that the element types get lost since we created an
        # untyped record (an anonymous composite type). For the same
        # reason this is also a normal tuple, not a named tuple.
        text_row = tuple(None if v is None else str(v) for v in value)
        self.assertEqual(row, text_row)

    def test_custom_type(self):
        values = [3, 5, 65]
        values = list(map(PgBitString, values))  # type: ignore
        table = self.table_prefix + 'booze'
        con = self._connect()
        try:
            cur = con.cursor()
            seq_params = enumerate(values)  # params have __pg_repr__ method
            cur.execute(
                f'create table "{table}" (n smallint, b bit varying(7))')
            cur.executemany(f"insert into {table} values (%s,%s)", seq_params)
            cur.execute(f"select * from {table}")
            rows = cur.fetchall()
        finally:
            con.close()
        self.assertEqual(len(rows), len(values))
        con = self._connect()
        try:
            cur = con.cursor()
            params = (1, object())  # an object that cannot be handled
            self.assertRaises(
                pgdb.InterfaceError, cur.execute,
                f"insert into {table} values (%s,%s)", params)
        finally:
            con.close()

    def test_set_decimal_type(self):
        from pgdb.cast import decimal_type
        self.assertIs(decimal_type(), Decimal)
        con = self._connect()
        try:
            cur = con.cursor()
            # change decimal type globally to int

            class CustomDecimal(str):

                def __init__(self, value: Any) -> None:
                    self.value = value

                def __str__(self) -> str:
                    return str(self.value).replace('.', ',')

            self.assertIs(decimal_type(CustomDecimal), CustomDecimal)
            cur.execute('select 4.25')
            self.assertEqual(cur.description[0].type_code, pgdb.NUMBER)
            value = cur.fetchone()[0]
            self.assertIsInstance(value, CustomDecimal)
            self.assertEqual(str(value), '4,25')
            # change decimal type again to float
            self.assertIs(decimal_type(float), float)
            cur.execute('select 4.25')
            self.assertEqual(cur.description[0].type_code, pgdb.NUMBER)
            value = cur.fetchone()[0]
            # the connection still uses the old setting
            self.assertIsInstance(value, str)
            self.assertEqual(str(value), '4,25')
            # bust the cache for type functions for the connection
            con.type_cache.reset_typecast()
            cur.execute('select 4.25')
            self.assertEqual(cur.description[0].type_code, pgdb.NUMBER)
            value = cur.fetchone()[0]
            # now the connection uses the new setting
            self.assertIsInstance(value, float)
            self.assertEqual(value, 4.25)
        finally:
            con.close()
            decimal_type(Decimal)
        self.assertIs(decimal_type(), Decimal)

    def test_global_typecast(self):
        try:
            query = 'select 2::int2, 4::int4, 8::int8'
            self.assertIs(pgdb.get_typecast('int4'), int)
            cast_int = lambda v: f'int({v})'  # noqa: E731
            pgdb.set_typecast('int4', cast_int)
            con = self._connect()
            try:
                i2, i4, i8 = con.cursor().execute(query).fetchone()
            finally:
                con.close()
            self.assertEqual(i2, 2)
            self.assertEqual(i4, 'int(4)')
            self.assertEqual(i8, 8)
            pgdb.set_typecast(['int2', 'int8'], cast_int)
            con = self._connect()
            try:
                i2, i4, i8 = con.cursor().execute(query).fetchone()
            finally:
                con.close()
            self.assertEqual(i2, 'int(2)')
            self.assertEqual(i4, 'int(4)')
            self.assertEqual(i8, 'int(8)')
            pgdb.reset_typecast('int4')
            con = self._connect()
            try:
                i2, i4, i8 = con.cursor().execute(query).fetchone()
            finally:
                con.close()
            self.assertEqual(i2, 'int(2)')
            self.assertEqual(i4, 4)
            self.assertEqual(i8, 'int(8)')
            pgdb.reset_typecast(['int2', 'int8'])
            con = self._connect()
            try:
                i2, i4, i8 = con.cursor().execute(query).fetchone()
            finally:
                con.close()
            self.assertEqual(i2, 2)
            self.assertEqual(i4, 4)
            self.assertEqual(i8, 8)
            pgdb.set_typecast(['int2', 'int8'], cast_int)
            con = self._connect()
            try:
                i2, i4, i8 = con.cursor().execute(query).fetchone()
            finally:
                con.close()
            self.assertEqual(i2, 'int(2)')
            self.assertEqual(i4, 4)
            self.assertEqual(i8, 'int(8)')
        finally:
            pgdb.reset_typecast()
        con = self._connect()
        try:
            i2, i4, i8 = con.cursor().execute(query).fetchone()
        finally:
            con.close()
        self.assertEqual(i2, 2)
        self.assertEqual(i4, 4)
        self.assertEqual(i8, 8)

    def test_set_typecast_for_arrays(self):
        query = 'select ARRAY[1,2,3]'
        try:
            con = self._connect()
            try:
                r = con.cursor().execute(query).fetchone()[0]
            finally:
                con.close()
            self.assertIsInstance(r, list)
            self.assertEqual(r, [1, 2, 3])
            pgdb.set_typecast('anyarray', lambda v, basecast: v)
            con = self._connect()
            try:
                r = con.cursor().execute(query).fetchone()[0]
            finally:
                con.close()
            self.assertIsInstance(r, str)
            self.assertEqual(r, '{1,2,3}')
        finally:
            pgdb.reset_typecast()
        con = self._connect()
        try:
            r = con.cursor().execute(query).fetchone()[0]
        finally:
            con.close()
        self.assertIsInstance(r, list)
        self.assertEqual(r, [1, 2, 3])

    def test_unicode_with_utf8(self):
        table = self.table_prefix + 'booze'
        s = "He wes Leovenaðes sone — liðe him be Drihten"
        con = self._connect()
        cur = con.cursor()
        try:
            cur.execute(f"create table {table} (t text)")
            try:
                cur.execute("set client_encoding=utf8")
                cur.execute(f"select '{s}'")
            except Exception:
                self.skipTest("database does not support utf8")
            output1 = cur.fetchone()[0]
            cur.execute(f"insert into {table} values (%s)", (s,))
            cur.execute(f"select * from {table}")
            output2 = cur.fetchone()[0]
            cur.execute(f"select t = '{s}' from {table}")
            output3 = cur.fetchone()[0]
            cur.execute(f"select t = %s from {table}", (s,))
            output4 = cur.fetchone()[0]
        finally:
            con.close()
        self.assertIsInstance(output1, str)
        self.assertEqual(output1, s)
        self.assertIsInstance(output2, str)
        self.assertEqual(output2, s)
        self.assertIsInstance(output3, bool)
        self.assertTrue(output3)
        self.assertIsInstance(output4, bool)
        self.assertTrue(output4)

    def test_unicode_with_latin1(self):
        table = self.table_prefix + 'booze'
        s = "Ehrt den König seine Würde, ehret uns der Hände Fleiß."
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute(f"create table {table} (t text)")
            try:
                cur.execute("set client_encoding=latin1")
                cur.execute(f"select '{s}'")
            except Exception:
                self.skipTest("database does not support latin1")
            output1 = cur.fetchone()[0]
            cur.execute(f"insert into {table} values (%s)", (s,))
            cur.execute(f"select * from {table}")
            output2 = cur.fetchone()[0]
            cur.execute(f"select t = '{s}' from {table}")
            output3 = cur.fetchone()[0]
            cur.execute(f"select t = %s from {table}", (s,))
            output4 = cur.fetchone()[0]
        finally:
            con.close()
        self.assertIsInstance(output1, str)
        self.assertEqual(output1, s)
        self.assertIsInstance(output2, str)
        self.assertEqual(output2, s)
        self.assertIsInstance(output3, bool)
        self.assertTrue(output3)
        self.assertIsInstance(output4, bool)
        self.assertTrue(output4)

    def test_bool(self):
        values = [False, True, None, 't', 'f', 'true', 'false']
        table = self.table_prefix + 'booze'
        con = self._connect()
        try:
            cur = con.cursor()
            cur.execute(f"create table {table} (n smallint, booltest bool)")
            params = enumerate(values)
            cur.executemany(f"insert into {table} values (%s,%s)", params)
            cur.execute(f"select booltest from {table} order by n")
            rows = cur.fetchall()
            self.assertEqual(cur.description[0].type_code, pgdb.BOOL)
        finally:
            con.close()
        rows = [row[0] for row in rows]
        values[3] = values[5] = True
        values[4] = values[6] = False
        self.assertEqual(rows, values)

    def test_literal(self):
        con = self._connect()
        try:
            cur = con.cursor()
            value = "lower('Hello')"
            cur.execute("select %s, %s", (value, pgdb.Literal(value)))
            row = cur.fetchone()
        finally:
            con.close()
        self.assertEqual(row, (value, 'hello'))

    def test_json(self):
        inval = {"employees": [
            {"firstName": "John", "lastName": "Doe", "age": 61}]}
        table = self.table_prefix + 'booze'
        con = self._connect()
        try:
            cur = con.cursor()
            try:
                cur.execute(f"create table {table} (jsontest json)")
            except pgdb.ProgrammingError:
                self.skipTest('database does not support json')
            params = (pgdb.Json(inval),)
            cur.execute(f"insert into {table} values (%s)", params)
            cur.execute(f"select jsontest from {table}")
            outval = cur.fetchone()[0]
            self.assertEqual(cur.description[0].type_code, pgdb.JSON)
        finally:
            con.close()
        self.assertEqual(inval, outval)

    def test_jsonb(self):
        inval = {"employees": [
            {"firstName": "John", "lastName": "Doe", "age": 61}]}
        table = self.table_prefix + 'booze'
        con = self._connect()
        try:
            cur = con.cursor()
            try:
                cur.execute(f"create table {table} (jsonbtest jsonb)")
            except pgdb.ProgrammingError:
                self.skipTest('database does not support jsonb')
            params = (pgdb.Json(inval),)
            cur.execute(f"insert into {table} values (%s)", params)
            cur.execute(f"select jsonbtest from {table}")
            outval = cur.fetchone()[0]
            self.assertEqual(cur.description[0].type_code, pgdb.JSON)
        finally:
            con.close()
        self.assertEqual(inval, outval)

    def test_execute_edge_cases(self):
        con = self._connect()
        try:
            cur = con.cursor()
            sql = 'invalid'  # should be ignored with empty parameter list
            cur.executemany(sql, [])
            sql = 'select %d + 1'
            cur.execute(sql, [(1,), (2,)])  # deprecated use of execute()
            self.assertEqual(cur.fetchone()[0], 3)
            sql = 'select 1/0'  # cannot be executed
            self.assertRaises(pgdb.DataError, cur.execute, sql)
            cur.close()
            con.rollback()
            if pgdb.shortcutmethods:
                res = con.execute('select %d', (1,)).fetchone()
                self.assertEqual(res, (1,))
                res = con.executemany('select %d', [(1,), (2,)]).fetchone()
                self.assertEqual(res, (2,))
        finally:
            con.close()
        sql = 'select 1'  # cannot be executed after connection is closed
        self.assertRaises(pgdb.OperationalError, cur.execute, sql)

    def test_fetchall_with_various_sizes(self):
        # we test this because there are optimizations based on result size
        con = self._connect()
        try:
            for n in (1, 3, 5, 7, 10, 100, 1000):
                cur = con.cursor()
                try:
                    cur.execute('select n, n::text as s, n % 2 = 1 as b'
                                f' from generate_series(1, {n}) as s(n)')
                    res = cur.fetchall()
                    self.assertEqual(len(res), n, res)
                    self.assertEqual(len(res[0]), 3)
                    self.assertEqual(res[0].n, 1)
                    self.assertEqual(res[0].s, '1')
                    self.assertIs(res[0].b, True)
                    self.assertEqual(len(res[-1]), 3)
                    self.assertEqual(res[-1].n, n)
                    self.assertEqual(res[-1].s, str(n))
                    self.assertIs(res[-1].b, n % 2 == 1)
                finally:
                    cur.close()
        finally:
            con.close()

    def test_fetchmany_with_keep(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.assertEqual(cur.arraysize, 1)
            cur.execute('select * from generate_series(1, 25)')
            self.assertEqual(len(cur.fetchmany()), 1)
            self.assertEqual(len(cur.fetchmany()), 1)
            self.assertEqual(cur.arraysize, 1)
            cur.arraysize = 3
            self.assertEqual(len(cur.fetchmany()), 3)
            self.assertEqual(len(cur.fetchmany()), 3)
            self.assertEqual(cur.arraysize, 3)
            self.assertEqual(len(cur.fetchmany(size=2)), 2)
            self.assertEqual(cur.arraysize, 3)
            self.assertEqual(len(cur.fetchmany()), 3)
            self.assertEqual(len(cur.fetchmany()), 3)
            self.assertEqual(len(cur.fetchmany(size=2, keep=True)), 2)
            self.assertEqual(cur.arraysize, 2)
            self.assertEqual(len(cur.fetchmany()), 2)
            self.assertEqual(len(cur.fetchmany()), 2)
            self.assertEqual(len(cur.fetchmany(25)), 3)
        finally:
            con.close()

    def help_nextset_setup(self, _cur):
        pass  # helper not needed

    def help_nextset_teardown(self, _cur):
        pass  # helper not needed

    def test_nextset(self):
        con = self._connect()
        cur = con.cursor()
        self.assertRaises(con.NotSupportedError, cur.nextset)

    def test_setoutputsize(self):
        pass  # not supported

    def test_connection_errors(self):
        con = self._connect()
        self.assertEqual(con.Error, pgdb.Error)
        self.assertEqual(con.Warning, pgdb.Warning)
        self.assertEqual(con.InterfaceError, pgdb.InterfaceError)
        self.assertEqual(con.DatabaseError, pgdb.DatabaseError)
        self.assertEqual(con.InternalError, pgdb.InternalError)
        self.assertEqual(con.OperationalError, pgdb.OperationalError)
        self.assertEqual(con.ProgrammingError, pgdb.ProgrammingError)
        self.assertEqual(con.IntegrityError, pgdb.IntegrityError)
        self.assertEqual(con.DataError, pgdb.DataError)
        self.assertEqual(con.NotSupportedError, pgdb.NotSupportedError)

    def test_transaction(self):
        table = self.table_prefix + 'booze'
        con1 = self._connect()
        cur1 = con1.cursor()
        self.execute_ddl1(cur1)
        con1.commit()
        con2 = self._connect()
        cur2 = con2.cursor()
        cur2.execute(f"select name from {table}")
        self.assertIsNone(cur2.fetchone())
        cur1.execute(f"insert into {table} values('Schlafly')")
        cur2.execute(f"select name from {table}")
        self.assertIsNone(cur2.fetchone())
        con1.commit()
        cur2.execute(f"select name from {table}")
        self.assertEqual(cur2.fetchone(), ('Schlafly',))
        con2.close()
        con1.close()

    def test_autocommit(self):
        table = self.table_prefix + 'booze'
        con1 = self._connect()
        con1.autocommit = True
        cur1 = con1.cursor()
        self.execute_ddl1(cur1)
        con2 = self._connect()
        cur2 = con2.cursor()
        cur2.execute(f"select name from {table}")
        self.assertIsNone(cur2.fetchone())
        cur1.execute(f"insert into {table} values('Shmaltz Pastrami')")
        cur2.execute(f"select name from {table}")
        self.assertEqual(cur2.fetchone(), ('Shmaltz Pastrami',))
        con2.close()
        con1.close()

    def test_connection_as_contextmanager(self):
        table = self.table_prefix + 'booze'
        for autocommit in False, True:
            con = self._connect()
            con.autocommit = autocommit
            try:
                cur = con.cursor()
                if autocommit:
                    cur.execute(f"truncate table {table}")
                else:
                    cur.execute(
                        f"create table {table} (n smallint check(n!=4))")
                with con:
                    cur.execute(f"insert into {table} values (1)")
                    cur.execute(f"insert into {table} values (2)")
                try:
                    with con:
                        cur.execute(f"insert into {table} values (3)")
                        cur.execute(f"insert into {table} values (4)")
                except con.IntegrityError as error:
                    self.assertIn('check', str(error).lower())
                with con:
                    cur.execute(f"insert into {table} values (5)")
                    cur.execute(f"insert into {table} values (6)")
                try:
                    with con:
                        cur.execute(f"insert into {table} values (7)")
                        cur.execute(f"insert into {table} values (8)")
                        raise ValueError('transaction should rollback')
                except ValueError as error:
                    self.assertEqual(str(error), 'transaction should rollback')
                with con:
                    cur.execute(f"insert into {table} values (9)")
                cur.execute(f"select * from {table} order by 1")
                rows = cur.fetchall()
                rows = [row[0] for row in rows]
            finally:
                con.close()
            self.assertEqual(rows, [1, 2, 5, 6, 9])

    def test_cursor_connection(self):
        con = self._connect()
        cur = con.cursor()
        self.assertEqual(cur.connection, con)
        cur.close()

    def test_cursor_as_contextmanager(self):
        con = self._connect()
        with con.cursor() as cur:
            self.assertEqual(cur.connection, con)

    def test_pgdb_type(self):
        self.assertEqual(pgdb.STRING, pgdb.STRING)
        self.assertNotEqual(pgdb.STRING, pgdb.INTEGER)
        self.assertNotEqual(pgdb.STRING, pgdb.BOOL)
        self.assertNotEqual(pgdb.BOOL, pgdb.INTEGER)
        self.assertEqual(pgdb.INTEGER, pgdb.INTEGER)
        self.assertNotEqual(pgdb.INTEGER, pgdb.NUMBER)
        self.assertEqual('char', pgdb.STRING)
        self.assertEqual('varchar', pgdb.STRING)
        self.assertEqual('text', pgdb.STRING)
        self.assertNotEqual('numeric', pgdb.STRING)
        self.assertEqual('numeric', pgdb.NUMERIC)
        self.assertEqual('numeric', pgdb.NUMBER)
        self.assertEqual('int4', pgdb.NUMBER)
        self.assertNotEqual('int4', pgdb.NUMERIC)
        self.assertEqual('int2', pgdb.SMALLINT)
        self.assertNotEqual('int4', pgdb.SMALLINT)
        self.assertEqual('int2', pgdb.INTEGER)
        self.assertEqual('int4', pgdb.INTEGER)
        self.assertEqual('int8', pgdb.INTEGER)
        self.assertNotEqual('int4', pgdb.LONG)
        self.assertEqual('int8', pgdb.LONG)
        self.assertIn('char', pgdb.STRING)
        self.assertLess(pgdb.NUMERIC, pgdb.NUMBER)
        self.assertGreaterEqual(pgdb.NUMBER, pgdb.INTEGER)
        self.assertLessEqual(pgdb.TIME, pgdb.DATETIME)
        self.assertGreaterEqual(pgdb.DATETIME, pgdb.DATE)
        self.assertEqual(pgdb.ARRAY, pgdb.ARRAY)
        self.assertNotEqual(pgdb.ARRAY, pgdb.STRING)
        self.assertEqual('_char', pgdb.ARRAY)
        self.assertNotEqual('char', pgdb.ARRAY)
        self.assertEqual(pgdb.RECORD, pgdb.RECORD)
        self.assertNotEqual(pgdb.RECORD, pgdb.STRING)
        self.assertNotEqual(pgdb.RECORD, pgdb.ARRAY)
        self.assertEqual('record', pgdb.RECORD)
        self.assertNotEqual('_record', pgdb.RECORD)

    def test_no_close(self):
        data = ('hello', 'world')
        con = self._connect()
        cur = con.cursor()
        cur.build_row_factory = lambda: tuple
        cur.execute("select %s, %s", data)
        row = cur.fetchone()
        self.assertEqual(row, data)

    def test_change_row_factory_cache_size(self):
        from pg import RowCache
        queries = ['select 1 as a, 2 as b, 3 as c', 'select 123 as abc']
        con = self._connect()
        cur = con.cursor()
        for maxsize in (None, 0, 1, 2, 3, 10, 1024):
            RowCache.change_size(maxsize)
            for _i in range(3):
                for q in queries:
                    cur.execute(q)
                    r = cur.fetchone()
                    if q.endswith('abc'):
                        self.assertEqual(r, (123,))
                        self.assertEqual(r._fields, ('abc',))
                    else:
                        self.assertEqual(r, (1, 2, 3))
                        self.assertEqual(r._fields, ('a', 'b', 'c'))
            info = RowCache.row_factory.cache_info()
            self.assertEqual(info.maxsize, maxsize)
            self.assertEqual(info.hits + info.misses, 6)
            self.assertEqual(info.hits,
                             0 if maxsize is not None and maxsize < 2 else 4)

    def test_memory_leaks(self):
        ids: set = set()
        objs: list = []
        add_ids = ids.update
        gc.collect()
        objs[:] = gc.get_objects()
        add_ids(id(obj) for obj in objs)
        self.test_no_close()
        gc.collect()
        objs[:] = gc.get_objects()
        objs[:] = [obj for obj in objs if id(obj) not in ids]
        self.assertEqual(len(objs), 0)

    def test_cve_2018_1058(self):
        # internal queries should use qualified table and operator names,
        # see https://nvd.nist.gov/vuln/detail/CVE-2018-1058
        con = self._connect()
        cur = con.cursor()
        execute = cur.execute
        try:
            execute("SET client_min_messages TO WARNING")
            execute("SET TIMEZONE TO 'UTC'")
            execute("SHOW TIMEZONE")
            self.assertEqual(cur.fetchone()[0], 'UTC')
            execute("""
                CREATE OR REPLACE FUNCTION public.bad_eq(oid, integer)
                RETURNS boolean AS $$
                BEGIN
                  SET TIMEZONE TO 'Europe/Athens';
                  RETURN oideq($1, $2::oid);
                END
                $$ LANGUAGE plpgsql
                """)
            execute("DROP OPERATOR IF EXISTS public.= (oid, integer)")
            execute("""
                CREATE OPERATOR public.= (
                  PROCEDURE = public.bad_eq,
                  LEFTARG = oid, RIGHTARG = integer
                );
                """)
            # the following select changes the time zone as a side effect if
            # internal query uses unqualified = operator as it did earlier
            execute("SELECT 1")
            execute("SHOW TIMEZONE")  # make sure time zone has not changed
            self.assertEqual(cur.fetchone()[0], 'UTC')
        finally:
            execute("DROP OPERATOR IF EXISTS public.= (oid, integer)")
            execute("DROP FUNCTION IF EXISTS public.bad_eq(oid, integer)")
            cur.close()
            con.close()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python

"""Test the classic PyGreSQL interface.

Sub-tests for large object support.

Contributed by Christoph Zwerschke.

These tests need a database to test against.
"""

import os
import tempfile
import unittest
from contextlib import suppress
from typing import Any

import pg  # the module under test

from .config import dbhost, dbname, dbpasswd, dbport, dbuser

windows = os.name == 'nt'


# noinspection PyArgumentList
def connect():
    """Create a basic pg connection to the test database."""
    connection = pg.connect(dbname, dbhost, dbport,
                            user=dbuser, passwd=dbpasswd)
    connection.query("set client_min_messages=warning")
    return connection


class TestModuleConstants(unittest.TestCase):
    """Test the existence of the documented module constants."""

    def test_large_object_int_constants(self):
        names = ['INV_READ', 'INV_WRITE', 'SEEK_SET', 'SEEK_CUR', 'SEEK_END']
        for name in names:
            try:
                value = getattr(pg, name)
            except AttributeError:
                self.fail(f'Module constant {name} is missing')
            self.assertIsInstance(value, int)


class TestCreatingLargeObjects(unittest.TestCase):
    """Test creating large objects using a connection."""

    def setUp(self):
        self.c = connect()
        self.c.query('begin')

    def tearDown(self):
        self.c.query('rollback')
        self.c.close()

    def assertIsLargeObject(self, obj):  # noqa: N802
        self.assertIsNotNone(obj)
        self.assertTrue(hasattr(obj, 'open'))
        self.assertTrue(hasattr(obj, 'close'))
        self.assertTrue(hasattr(obj, 'oid'))
        self.assertTrue(hasattr(obj, 'pgcnx'))
        self.assertTrue(hasattr(obj, 'error'))
        self.assertIsInstance(obj.oid, int)
        self.assertNotEqual(obj.oid, 0)
        self.assertIs(obj.pgcnx, self.c)
        self.assertIsInstance(obj.error, str)
        self.assertFalse(obj.error)

    def test_lo_create(self):
        large_object = self.c.locreate(pg.INV_READ | pg.INV_WRITE)
        try:
            self.assertIsLargeObject(large_object)
        finally:
            del large_object

    def test_get_lo(self):
        large_object = self.c.locreate(pg.INV_READ | pg.INV_WRITE)
        try:
            self.assertIsLargeObject(large_object)
            oid = large_object.oid
        finally:
            del large_object
        data = b'some data to be shared'
        large_object = self.c.getlo(oid)
        try:
            self.assertIsLargeObject(large_object)
            self.assertEqual(large_object.oid, oid)
            large_object.open(pg.INV_WRITE)
            large_object.write(data)
            large_object.close()
        finally:
            del large_object
        large_object = self.c.getlo(oid)
        try:
            self.assertIsLargeObject(large_object)
            self.assertEqual(large_object.oid, oid)
            large_object.open(pg.INV_READ)
            r = large_object.read(80)
            large_object.close()
            large_object.unlink()
        finally:
            del large_object
        self.assertIsInstance(r, bytes)
        self.assertEqual(r, data)

    def test_lo_import(self):
        f : Any
        if windows:
            # NamedTemporaryFiles don't work well here
            fname = 'temp_test_pg_largeobj_import.txt'
            f = open(fname, 'wb')  # noqa: SIM115
        else:
            f = tempfile.NamedTemporaryFile()  # noqa: SIM115
            fname = f.name
        data = b'some data to be imported'
        f.write(data)
        if windows:
            f.close()
            f = open(fname, 'rb')  # noqa: SIM115
        else:
            f.flush()
            f.seek(0)
        large_object = self.c.loimport(f.name)
        try:
            f.close()
            if windows:
                os.remove(fname)
            self.assertIsLargeObject(large_object)
            large_object.open(pg.INV_READ)
            large_object.seek(0, pg.SEEK_SET)
            r = large_object.size()
            self.assertIsInstance(r, int)
            self.assertEqual(r, len(data))
            r = large_object.read(80)
            self.assertIsInstance(r, bytes)
            self.assertEqual(r, data)
            large_object.close()
            large_object.unlink()
        finally:
            del large_object


class TestLargeObjects(unittest.TestCase):
    """Test the large object methods."""

    def setUp(self):
        self.pgcnx = connect()
        self.pgcnx.query('begin')
        self.obj = self.pgcnx.locreate(pg.INV_READ | pg.INV_WRITE)

    def tearDown(self):
        if self.obj.oid:
            with suppress(SystemError, OSError):
                self.obj.close()
            with suppress(SystemError, OSError):
                self.obj.unlink()
        del self.obj
        with suppress(SystemError):
            self.pgcnx.query('rollback')
        self.pgcnx.close()

    def test_class_name(self):
        self.assertEqual(self.obj.__class__.__name__, 'LargeObject')

    def test_module_name(self):
        self.assertEqual(self.obj.__class__.__module__, 'pg')

    def test_oid(self):
        self.assertIsInstance(self.obj.oid, int)
        self.assertNotEqual(self.obj.oid, 0)

    def test_pgcn(self):
        self.assertIs(self.obj.pgcnx, self.pgcnx)

    def test_error(self):
        self.assertIsInstance(self.obj.error, str)
        self.assertEqual(self.obj.error, '')

    def test_str(self):
        self.obj.open(pg.INV_WRITE)
        data = b'some object to be printed'
        self.obj.write(data)
        oid = self.obj.oid
        r = str(self.obj)
        self.assertEqual(r, f'Opened large object, oid {oid}')
        self.obj.close()
        r = str(self.obj)
        self.assertEqual(r, f'Closed large object, oid {oid}')

    def test_repr(self):
        r = repr(self.obj)
        self.assertTrue(r.startswith('<pg.LargeObject object'), r)

    def test_open(self):
        # noinspection PyShadowingBuiltins
        open = self.obj.open
        # testing with invalid parameters
        self.assertRaises(TypeError, open)
        self.assertRaises(TypeError, open, pg.INV_READ, pg.INV_WRITE)
        open(pg.INV_READ)
        # object is already open
        self.assertRaises(IOError, open, pg.INV_READ)

    def test_close(self):
        close = self.obj.close
        # testing with invalid parameters
        self.assertRaises(TypeError, close, pg.INV_READ)
        # object is not yet open
        self.assertRaises(IOError, close)
        self.obj.open(pg.INV_READ)
        close()
        self.assertRaises(IOError, close)

    def test_read(self):
        read = self.obj.read
        # testing with invalid parameters
        self.assertRaises(TypeError, read)
        self.assertRaises(ValueError, read, -1)
        self.assertRaises(TypeError, read, 'invalid')
        self.assertRaises(TypeError, read, 80, 'invalid')
        # reading when object is not yet open
        self.assertRaises(IOError, read, 80)
        data = b'some data to be read'
        self.obj.open(pg.INV_WRITE)
        self.obj.write(data)
        self.obj.close()
        self.obj.open(pg.INV_READ)
        r = read(80)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r, data)
        self.obj.close()
        self.obj.open(pg.INV_READ)
        r = read(8)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r, data[:8])
        self.obj.close()

    def test_write(self):
        write = self.obj.write
        # testing with invalid parameters
        self.assertRaises(TypeError, write)
        self.assertRaises(TypeError, write, -1)
        self.assertRaises(TypeError, write, '', 'invalid')
        # writing when object is not yet open
        self.assertRaises(IOError, write, 'invalid')
        data = b'some data to be written'
        self.obj.open(pg.INV_WRITE)
        write(data)
        self.obj.close()
        self.obj.open(pg.INV_READ)
        r = self.obj.read(80)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r, data)

    def test_write_latin1_bytes(self):
        read = self.obj.read
        self.obj.open(pg.INV_WRITE)
        self.obj.write('käse'.encode('latin1'))
        self.obj.close()
        self.obj.open(pg.INV_READ)
        r = read(80)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r.decode('latin1'), 'käse')

    def test_write_utf8_bytes(self):
        read = self.obj.read
        self.obj.open(pg.INV_WRITE)
        self.obj.write('käse'.encode())
        self.obj.close()
        self.obj.open(pg.INV_READ)
        r = read(80)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r.decode(), 'käse')

    def test_write_utf8_string(self):
        read = self.obj.read
        self.obj.open(pg.INV_WRITE)
        self.obj.write('käse')
        self.obj.close()
        self.obj.open(pg.INV_READ)
        r = read(80)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r.decode(), 'käse')

    def test_seek(self):
        seek = self.obj.seek
        # testing with invalid parameters
        self.assertRaises(TypeError, seek)
        self.assertRaises(TypeError, seek, 0)
        self.assertRaises(TypeError, seek, 0, pg.SEEK_SET, pg.SEEK_END)
        self.assertRaises(TypeError, seek, 'invalid', pg.SEEK_SET)
        self.assertRaises(TypeError, seek, 0, 'invalid')
        # seeking when object is not yet open
        self.assertRaises(IOError, seek, 0, pg.SEEK_SET)
        data = b'some data to be seeked'
        self.obj.open(pg.INV_WRITE)
        self.obj.write(data)
        self.obj.close()
        self.obj.open(pg.INV_READ)
        seek(0, pg.SEEK_SET)
        r = self.obj.read(9)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r, b'some data')
        seek(4, pg.SEEK_CUR)
        r = self.obj.read(2)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r, b'be')
        seek(-10, pg.SEEK_CUR)
        r = self.obj.read(4)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r, b'data')
        seek(0, pg.SEEK_SET)
        r = self.obj.read(4)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r, b'some')
        seek(-6, pg.SEEK_END)
        r = self.obj.read(4)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r, b'seek')

    def test_tell(self):
        tell = self.obj.tell
        # testing with invalid parameters
        self.assertRaises(TypeError, tell, 0)
        # telling when object is not yet open
        self.assertRaises(IOError, tell)
        data = b'some story to be told'
        self.obj.open(pg.INV_WRITE)
        self.obj.write(data)
        r = tell()
        self.assertIsInstance(r, int)
        self.assertEqual(r, len(data))
        self.obj.close()
        self.obj.open(pg.INV_READ)
        r = tell()
        self.assertIsInstance(r, int)
        self.assertEqual(r, 0)
        self.obj.seek(5, pg.SEEK_SET)
        r = tell()
        self.assertIsInstance(r, int)
        self.assertEqual(r, 5)

    def test_unlink(self):
        unlink = self.obj.unlink
        # testing with invalid parameters
        self.assertRaises(TypeError, unlink, 0)
        # unlinking when object is still open
        self.obj.open(pg.INV_WRITE)
        self.assertIsNotNone(self.obj.oid)
        self.assertNotEqual(0, self.obj.oid)
        self.assertRaises(IOError, unlink)
        data = b'some data to be sold'
        self.obj.write(data)
        self.obj.close()
        # unlinking after object has been closed
        unlink()
        self.assertIsNone(self.obj.oid)
        # unlinking after object has been already unlinked
        self.assertRaises(pg.IntegrityError, unlink)

    def test_unlink_inexistent(self):
        unlink = self.obj.unlink
        self.obj.open(pg.INV_WRITE)
        self.obj.close()
        self.pgcnx.query(f'select lo_unlink({self.obj.oid})')
        self.assertRaises(IOError, unlink)

    def test_size(self):
        size = self.obj.size
        # testing with invalid parameters
        self.assertRaises(TypeError, size, 0)
        # sizing when object is not yet open
        self.assertRaises(IOError, size)
        # sizing an empty object
        self.obj.open(pg.INV_READ)
        r = size()
        self.obj.close()
        self.assertIsInstance(r, int)
        self.assertEqual(r, 0)
        # sizing after adding some data
        data = b'some data to be sized'
        self.obj.open(pg.INV_WRITE)
        self.obj.write(data)
        self.obj.close()
        # sizing when current position is zero
        self.obj.open(pg.INV_READ)
        r = size()
        self.obj.close()
        self.assertIsInstance(r, int)
        self.assertEqual(r, len(data))
        self.obj.open(pg.INV_READ)
        # sizing when current position is not zero
        self.obj.seek(5, pg.SEEK_SET)
        r = size()
        self.obj.close()
        self.assertIsInstance(r, int)
        self.assertEqual(r, len(data))
        # sizing after adding more data
        data += b' and more data'
        self.obj.open(pg.INV_WRITE)
        self.obj.write(data)
        self.obj.close()
        self.obj.open(pg.INV_READ)
        r = size()
        self.obj.close()
        self.assertIsInstance(r, int)
        self.assertEqual(r, len(data))

    def test_export(self):
        export = self.obj.export
        # testing with invalid parameters
        self.assertRaises(TypeError, export)
        self.assertRaises(TypeError, export, 0)
        self.assertRaises(TypeError, export, 'invalid', 0)
        f: Any
        if windows:
            # NamedTemporaryFiles don't work well here
            fname = 'temp_test_pg_largeobj_export.txt'
            f = open(fname, 'wb')  # noqa: SIM115
        else:
            f = tempfile.NamedTemporaryFile()  # noqa: SIM115
            fname = f.name
        data = b'some data to be exported'
        self.obj.open(pg.INV_WRITE)
        self.obj.write(data)
        # exporting when object is not yet closed
        self.assertRaises(IOError, export, f.name)
        self.obj.close()
        export(fname)
        if windows:
            f.close()
            f = open(fname, 'rb')  # noqa: SIM115
        r = f.read()
        f.close()
        if windows:
            os.remove(fname)
        self.assertIsInstance(r, bytes)
        self.assertEqual(r, data)

    def test_export_in_existent(self):
        export = self.obj.export
        with tempfile.NamedTemporaryFile() as f:
            self.obj.open(pg.INV_WRITE)
            self.obj.close()
            self.pgcnx.query(f'select lo_unlink({self.obj.oid})')
            self.assertRaises(IOError, export, f.name)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python

"""Driver script for building PyGreSQL using setuptools.

You can build the PyGreSQL distribution like this:

    pip install build
    python -m build -C strict -C memory-size
"""

import contextlib
import os
import platform
import re
import sys
import sysconfig
import warnings

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

min_py_version = 3, 8  # supported: Python >= 3.8
max_py_version = 4, 0  # and < 4.0
min_pg_version = 12, 0  # supported: PostgreSQL >= 12.0


def project_version():
    """Read the PyGreSQL version from the pyproject.toml file."""
    with open('pyproject.toml') as f:
        for d in f:
            if d.startswith("version ="):
                version = d.split("=")[1].strip().strip('"')
                return version
    raise Exception("Cannot determine PyGreSQL version")


def project_readme():
    """Get the content of the README file."""
    with open('README.rst') as f:
        return f.read()


version = project_version()

if not min_py_version <= sys.version_info[:2] < max_py_version:
    raise Exception(
        f"Sorry, PyGreSQL {version} does not support this Python version")


def patch_pyproject_toml():
    """Patch pyproject.toml to make it work with old setuptools versions.
    
    This allows building PyGreSQL with Python < 3.9 which only supports
    setuptools up to version 75, since our pyproject.toml requires version 77.
    """
    from setuptools import __version__ as version

    try:
        version = int(version.split('.', 1)[0])
    except (IndexError, TypeError, ValueError):
        return
    if not 61 <= version < 77:  # only needed for setuptools 61 to 76
        return

    try:
        from setuptools.config import pyprojecttoml
    except ImportError:
        return

    load_file = pyprojecttoml.load_file

    def load_file_patched(filepath):
        d = load_file(filepath)
        with contextlib.suppress(KeyError):
            p = d['project']
            t = p['license']
            f = p.pop('license-files')
            p['license'] = {'text': t, 'files': f[0]}
        return d

    pyprojecttoml.load_file = load_file_patched


patch_pyproject_toml()  # needed for Python < 3.9


# For historical reasons, PyGreSQL does not install itself as a single
# "pygresql" package, but as two top-level modules "pg", providing the
# classic interface, and "pgdb" for the modern DB-API 2.0 interface.
# These two top-level Python modules share the same C extension "_pg".

def pg_config(s):
    """Retrieve information about installed version of PostgreSQL."""
    f = os.popen(f'pg_config --{s}')  # noqa: S605
    d = f.readline().strip()
    if f.close() is not None:
        raise Exception("pg_config tool is not available.")
    if not d:
        raise Exception(f"Could not get {s} information.")
    return d


def pg_version():
    """Return the PostgreSQL version as a tuple of integers."""
    match = re.search(r'(\d+)\.(\d+)', pg_config('version'))
    if match:
        return tuple(map(int, match.groups()))
    return min_pg_version


pg_version = pg_version()
libraries = ['pq']
# Make sure that the Python header files are searched before
# those of PostgreSQL, because PostgreSQL can have its own Python.h
include_dirs = [sysconfig.get_path("include"), pg_config('includedir')]
library_dirs = [sysconfig.get_path("purelib"), pg_config('libdir')]
define_macros = [('PYGRESQL_VERSION', version)]
undef_macros = []
extra_compile_args = ['-O2', '-funsigned-char', '-Wall', '-Wconversion']


class BuildPgExt(build_ext):
    """Customized build_ext command for PyGreSQL."""

    description = "build the PyGreSQL C extension"

    user_options = [*build_ext.user_options,  # noqa: RUF012
        ('strict', None, "count all compiler warnings as errors"),
        ('memory-size', None, "enable memory size function"),
        ('no-memory-size', None, "disable memory size function")]

    boolean_options = [*build_ext.boolean_options,  # noqa: RUF012
        'strict', 'memory-size']

    negative_opt = {  # noqa: RUF012
        'no-memory-size': 'memory-size'}

    def initialize_options(self):
        """Initialize the supported options with default values."""
        super().initialize_options()
        self.strict = False
        self.memory_size = None
        supported = pg_version >= min_pg_version
        if not supported:
            warnings.warn(
                "PyGreSQL does not support the installed PostgreSQL version.",
                stacklevel=2)

    def finalize_options(self):
        """Set values for all build_pg options.
        
        Some values are set in build_extensions() since they depend
        on the compiler version which is not yet known at this point.
        """
        super().finalize_options()
        if self.strict:
            extra_compile_args.append('-Werror')
        wanted = self.memory_size
        supported = pg_version >= min_pg_version
        if (wanted is None and supported) or wanted:
            define_macros.append(('MEMORY_SIZE', None))
            if not supported:
                warnings.warn(
                    "The installed PostgreSQL version"
                    " does not support the memory size function.",
                    stacklevel=2)

    def build_extensions(self):
        """Build the PyGreSQL C extension."""
        # Adjust settings for Windows platforms
        if sys.platform == 'win32':
            libraries[0] = 'lib' + libraries[0]
            if os.path.exists(os.path.join(
                    library_dirs[1], libraries[0] + 'dll.lib')):
                libraries[0] += 'dll'
            compiler_type = self.compiler.compiler_type
            if compiler_type == 'mingw32':  # MinGW
                if platform.architecture()[0] == '64bit':  # needs MinGW-w64
                    define_macros.append(('MS_WIN64', None))
            elif compiler_type == 'msvc':  # Microsoft Visual C++
                extra_compile_args[1:] = [
                    '-J', '-W3', '-WX', '-wd4391',
                    '-Dinline=__inline']  # needed for MSVC 9
        super().build_extensions()


setup(
    name='PyGreSQL',
    version=version,
    description='Python PostgreSQL Interfaces',
    long_description=project_readme(),
    long_description_content_type='text/x-rst',
    keywords='pygresql postgresql database api dbapi',
    author="D'Arcy J. M. Cain",
    author_email="darcy@PyGreSQL.org",
    url='https://pygresql.github.io/',
    download_url='https://pygresql.github.io/download/',
    project_urls={
        'Documentation': 'https://pygresql.github.io/contents/',
        'Issue Tracker': 'https://github.com/PyGreSQL/PyGreSQL/issues/',
        'Mailing List': 'https://mail.vex.net/mailman/listinfo/pygresql',
        'Source Code': 'https://github.com/PyGreSQL/PyGreSQL'},
    classifiers=[
        'Development Status :: 6 - Mature',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: C',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Programming Language :: SQL',
        'Topic :: Database',
        'Topic :: Database :: Front-Ends',
        'Topic :: Software Development :: Libraries :: Python Modules'],
    license='PostgreSQL',
    license_files=['LICENSE.txt'],
    zip_safe=False,
    packages=["pg", "pgdb"],
    package_data={"pg": ["py.typed"], "pgdb": ["py.typed"]},
    ext_modules=[Extension(
        'pg._pg', [
            "ext/pgmodule.c",
            "ext/pginternal.c",
            "ext/pgconn.c",
            "ext/pgquery.c",
            "ext/pgsource.c",
            "ext/pgnotice.c",
            "ext/pglarge.c",
        ],
        include_dirs=include_dirs, library_dirs=library_dirs,
        define_macros=define_macros, undef_macros=undef_macros,
        libraries=libraries, extra_compile_args=extra_compile_args)],
    cmdclass=dict(build_ext=BuildPgExt),
)



#!/usr/bin/python

"""Python DB API 2.0 driver compliance unit test suite.

This software is Public Domain and may be used without restrictions.

Some modernization of the code has been done by the PyGreSQL team.
"""

from __future__ import annotations

import time
import unittest
from contextlib import suppress
from typing import Any, ClassVar

__version__ = '1.15.0'

class DatabaseAPI20Test(unittest.TestCase):
    """Test a database self.driver for DB API 2.0 compatibility.

    This implementation tests Gadfly, but the TestCase
    is structured so that other self.drivers can subclass this
    test case to ensure compliance with the DB-API. It is
    expected that this TestCase may be expanded i  qn the future
    if ambiguities or edge conditions are discovered.

    The 'Optional Extensions' are not yet being tested.

    self.drivers should subclass this test, overriding setUp, tearDown,
    self.driver, connect_args and connect_kw_args. Class specification
    should be as follows:

    import dbapi20
    class mytest(dbapi20.DatabaseAPI20Test):
        [...]

    Don't 'import DatabaseAPI20Test from dbapi20', or you will
    confuse the unit tester - just 'import dbapi20'.
    """

    # The self.driver module. This should be the module where the 'connect'
    # method is to be found
    driver: Any = None
    connect_args: tuple = ()  # List of arguments to pass to connect
    connect_kw_args: ClassVar[dict[str, Any]] = {}  # Keyword arguments
    table_prefix = 'dbapi20test_'  # If you need to specify a prefix for tables

    ddl1 = f'create table {table_prefix}booze (name varchar(20))'
    ddl2 = (f'create table {table_prefix}barflys (name varchar(20),'
            ' drink varchar(30))')
    xddl1 = f'drop table {table_prefix}booze'
    xddl2 = f'drop table {table_prefix}barflys'
    insert = 'insert'

    lowerfunc = 'lower'  # Name of stored procedure to convert str to lowercase

    # Some drivers may need to override these helpers, for example adding
    # a 'commit' after the execute.
    def execute_ddl1(self, cursor):
        cursor.execute(self.ddl1)

    def execute_ddl2(self, cursor):
        cursor.execute(self.ddl2)

    def setUp(self):
        """Set up test fixture.

        self.drivers should override this method to perform required setup
        if any is necessary, such as creating the database.
        """
        pass

    def tearDown(self):
        """Tear down test fixture.

        self.drivers should override this method to perform required cleanup
        if any is necessary, such as deleting the test database.
        The default drops the tables that may be created.
        """
        try:
            con = self._connect()
            try:
                cur = con.cursor()
                for ddl in (self.xddl1, self.xddl2):
                    try:
                        cur.execute(ddl)
                        con.commit()
                    except self.driver.Error:
                        # Assume table didn't exist. Other tests will check if
                        # execute is busted.
                        pass
            finally:
                con.close()
        except Exception:
            pass

    def _connect(self):
        try:
            con = self.driver.connect(
                *self.connect_args, **self.connect_kw_args)
        except AttributeError:
            self.fail("No connect method found in self.driver module")
        if not isinstance(con, self.driver.Connection):
            self.fail("The connect method does not return a Connection")
        return con

    def test_connect(self):
        con = self._connect()
        con.close()

    def test_apilevel(self):
        try:
            # Must exist
            apilevel = self.driver.apilevel
            # Must equal 2.0
            self.assertEqual(apilevel, '2.0')
        except AttributeError:
            self.fail("Driver doesn't define apilevel")

    def test_threadsafety(self):
        try:
            # Must exist
            threadsafety = self.driver.threadsafety
            # Must be a valid value
            self.assertIn(threadsafety, (0, 1, 2, 3))
        except AttributeError:
            self.fail("Driver doesn't define threadsafety")

    def test_paramstyle(self):
        try:
            # Must exist
            paramstyle = self.driver.paramstyle
            # Must be a valid value
            self.assertIn(paramstyle, (
                'qmark', 'numeric', 'named', 'format', 'pyformat'))
        except AttributeError:
            self.fail("Driver doesn't define paramstyle")

    def test_exceptions(self):
        # Make sure required exceptions exist, and are in the
        # defined hierarchy.
        sub = issubclass
        self.assertTrue(sub(self.driver.Warning, Exception))
        self.assertTrue(sub(self.driver.Error, Exception))

        self.assertTrue(sub(self.driver.InterfaceError, self.driver.Error))
        self.assertTrue(sub(self.driver.DatabaseError, self.driver.Error))
        self.assertTrue(sub(self.driver.OperationalError, self.driver.Error))
        self.assertTrue(sub(self.driver.IntegrityError, self.driver.Error))
        self.assertTrue(sub(self.driver.InternalError, self.driver.Error))
        self.assertTrue(sub(self.driver.ProgrammingError, self.driver.Error))
        self.assertTrue(sub(self.driver.NotSupportedError, self.driver.Error))

    def test_exceptions_as_connection_attributes(self):
        # OPTIONAL EXTENSION
        # Test for the optional DB API 2.0 extension, where the exceptions
        # are exposed as attributes on the Connection object
        # I figure this optional extension will be implemented by any
        # driver author who is using this test suite, so it is enabled
        # by default.
        con = self._connect()
        drv = self.driver
        self.assertIs(con.Warning, drv.Warning)
        self.assertIs(con.Error, drv.Error)
        self.assertIs(con.InterfaceError, drv.InterfaceError)
        self.assertIs(con.DatabaseError, drv.DatabaseError)
        self.assertIs(con.OperationalError, drv.OperationalError)
        self.assertIs(con.IntegrityError, drv.IntegrityError)
        self.assertIs(con.InternalError, drv.InternalError)
        self.assertIs(con.ProgrammingError, drv.ProgrammingError)
        self.assertIs(con.NotSupportedError, drv.NotSupportedError)

    def test_commit(self):
        con = self._connect()
        try:
            # Commit must work, even if it doesn't do anything
            con.commit()
        finally:
            con.close()

    def test_rollback(self):
        con = self._connect()
        # If rollback is defined, it should either work or throw
        # the documented exception
        if hasattr(con, 'rollback'):
            with suppress(self.driver.NotSupportedError):
                # noinspection PyCallingNonCallable
                con.rollback()

    def test_cursor(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.assertIsNotNone(cur)
        finally:
            con.close()

    def test_cursor_isolation(self):
        con = self._connect()
        try:
            # Make sure cursors created from the same connection have
            # the documented transaction isolation level
            cur1 = con.cursor()
            cur2 = con.cursor()
            self.execute_ddl1(cur1)
            cur1.execute(f"{self.insert} into {self.table_prefix}booze"
                         " values ('Victoria Bitter')")
            cur2.execute(f"select name from {self.table_prefix}booze")
            booze = cur2.fetchall()
            self.assertEqual(len(booze), 1)
            self.assertEqual(len(booze[0]), 1)
            self.assertEqual(booze[0][0], 'Victoria Bitter')
        finally:
            con.close()

    def test_description(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.execute_ddl1(cur)
            self.assertIsNone(
                cur.description,
                'cursor.description should be none after executing a'
                ' statement that can return no rows (such as DDL)')
            cur.execute(f'select name from {self.table_prefix}booze')
            self.assertEqual(
                len(cur.description), 1,
                'cursor.description describes too many columns')
            self.assertEqual(
                len(cur.description[0]), 7,
                'cursor.description[x] tuples must have 7 elements')
            self.assertEqual(
                cur.description[0][0].lower(), 'name',
                'cursor.description[x][0] must return column name')
            self.assertEqual(
                cur.description[0][1], self.driver.STRING,
                'cursor.description[x][1] must return column type.'
                f' Got: {cur.description[0][1]!r}')

            # Make sure self.description gets reset
            self.execute_ddl2(cur)
            self.assertIsNone(
                cur.description,
                'cursor.description not being set to None when executing'
                ' no-result statements (eg. DDL)')
        finally:
            con.close()

    def test_rowcount(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.execute_ddl1(cur)
            self.assertIn(
                cur.rowcount, (-1, 0),  # Bug #543885
                'cursor.rowcount should be -1 or 0 after executing no-result'
                ' statements')
            cur.execute(f"{self.insert} into {self.table_prefix}booze"
                        " values ('Victoria Bitter')")
            self.assertIn(
                cur.rowcount, (-1, 1),
                'cursor.rowcount should == number or rows inserted, or'
                ' set to -1 after executing an insert statement')
            cur.execute(f"select name from {self.table_prefix}booze")
            self.assertIn(
                cur.rowcount, (-1, 1),
                'cursor.rowcount should == number of rows returned, or'
                ' set to -1 after executing a select statement')
            self.execute_ddl2(cur)
            self.assertIn(
                cur.rowcount, (-1, 0),  # Bug #543885
                'cursor.rowcount should be -1 or 0 after executing no-result'
                ' statements')
        finally:
            con.close()

    lower_func = 'lower'

    def test_callproc(self):
        con = self._connect()
        try:
            cur = con.cursor()
            if self.lower_func and hasattr(cur, 'callproc'):
                # noinspection PyCallingNonCallable
                r = cur.callproc(self.lower_func, ('FOO',))
                self.assertEqual(len(r), 1)
                self.assertEqual(r[0], 'FOO')
                r = cur.fetchall()
                self.assertEqual(len(r), 1, 'callproc produced no result set')
                self.assertEqual(
                    len(r[0]), 1, 'callproc produced invalid result set')
                self.assertEqual(
                    r[0][0], 'foo', 'callproc produced invalid results')
        finally:
            con.close()

    def test_close(self):
        con = self._connect()
        try:
            cur = con.cursor()
        finally:
            con.close()

        # cursor.execute should raise an Error if called after connection
        # closed
        self.assertRaises(self.driver.Error, self.execute_ddl1, cur)

        # connection.commit should raise an Error if called after connection'
        # closed.'
        self.assertRaises(self.driver.Error, con.commit)

    def test_non_idempotent_close(self):
        con = self._connect()
        con.close()
        # connection.close should raise an Error if called more than once
        # (the usefulness of this test and this feature is questionable)
        self.assertRaises(self.driver.Error, con.close)

    def test_execute(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self._paraminsert(cur)
        finally:
            con.close()

    def _paraminsert(self, cur):
        self.execute_ddl2(cur)
        table_prefix = self.table_prefix
        insert = f"{self.insert} into {table_prefix}barflys values"
        cur.execute(
            f"{insert} ('Victoria Bitter',"
            " 'thi%s :may ca%(u)se? troub:1e')")
        self.assertIn(cur.rowcount, (-1, 1))

        if self.driver.paramstyle == 'qmark':
            cur.execute(
                f"{insert} (?, 'thi%s :may ca%(u)se? troub:1e')",
                ("Cooper's",))
        elif self.driver.paramstyle == 'numeric':
            cur.execute(
                f"{insert} (:1, 'thi%s :may ca%(u)se? troub:1e')",
                ("Cooper's",))
        elif self.driver.paramstyle == 'named':
            cur.execute(
                f"{insert} (:beer, 'thi%s :may ca%(u)se? troub:1e')",
                {'beer': "Cooper's"})
        elif self.driver.paramstyle == 'format':
            cur.execute(
                f"{insert} (%s, 'thi%%s :may ca%%(u)se? troub:1e')",
                ("Cooper's",))
        elif self.driver.paramstyle == 'pyformat':
            cur.execute(
                f"{insert} (%(beer)s, 'thi%%s :may ca%%(u)se? troub:1e')",
                {'beer': "Cooper's"})
        else:
            self.fail('Invalid paramstyle')
        self.assertIn(cur.rowcount, (-1, 1))

        cur.execute(f'select name, drink from {table_prefix}barflys')
        res = cur.fetchall()
        self.assertEqual(len(res), 2, 'cursor.fetchall returned too few rows')
        beers = [res[0][0], res[1][0]]
        beers.sort()
        self.assertEqual(
            beers[0], "Cooper's",
            'cursor.fetchall retrieved incorrect data, or data inserted'
            ' incorrectly')
        self.assertEqual(
            beers[1], "Victoria Bitter",
            'cursor.fetchall retrieved incorrect data, or data inserted'
            ' incorrectly')
        trouble = "thi%s :may ca%(u)se? troub:1e"
        self.assertEqual(
            res[0][1], trouble,
            'cursor.fetchall retrieved incorrect data, or data inserted'
            f' incorrectly. Got: {res[0][1]!r}, Expected: {trouble!r}')
        self.assertEqual(
            res[1][1], trouble,
            'cursor.fetchall retrieved incorrect data, or data inserted'
            f' incorrectly. Got: {res[1][1]!r}, Expected: {trouble!r}')

    def test_executemany(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.execute_ddl1(cur)
            table_prefix = self.table_prefix
            insert = f'{self.insert} into {table_prefix}booze values'
            largs = [("Cooper's",), ("Boag's",)]
            margs = [{'beer': "Cooper's"}, {'beer': "Boag's"}]
            if self.driver.paramstyle == 'qmark':
                cur.executemany(f'{insert} (?)', largs)
            elif self.driver.paramstyle == 'numeric':
                cur.executemany(f'{insert} (:1)', largs)
            elif self.driver.paramstyle == 'named':
                cur.executemany(f'{insert} (:beer)', margs)
            elif self.driver.paramstyle == 'format':
                cur.executemany(f'{insert} (%s)', largs)
            elif self.driver.paramstyle == 'pyformat':
                cur.executemany(f'{insert} (%(beer)s)', margs)
            else:
                self.fail('Unknown paramstyle')
            self.assertIn(
                cur.rowcount, (-1, 2),
                'insert using cursor.executemany set cursor.rowcount to'
                f' incorrect value {cur.rowcount!r}')
            cur.execute(f'select name from {table_prefix}booze')
            res = cur.fetchall()
            self.assertEqual(
                len(res), 2,
                'cursor.fetchall retrieved incorrect number of rows')
            beers = [res[0][0], res[1][0]]
            beers.sort()
            self.assertEqual(beers[0], "Boag's", 'incorrect data retrieved')
            self.assertEqual(beers[1], "Cooper's", 'incorrect data retrieved')
        finally:
            con.close()

    def test_fetchone(self):
        con = self._connect()
        try:
            cur = con.cursor()

            # cursor.fetchone should raise an Error if called before
            # executing a select-type query
            self.assertRaises(self.driver.Error, cur.fetchone)

            # cursor.fetchone should raise an Error if called after
            # executing a query that cannot return rows
            self.execute_ddl1(cur)
            self.assertRaises(self.driver.Error, cur.fetchone)

            cur.execute(f'select name from {self.table_prefix}booze')
            self.assertIsNone(
                cur.fetchone(),
                'cursor.fetchone should return None if a query retrieves'
                ' no rows')
            self.assertIn(cur.rowcount, (-1, 0))

            # cursor.fetchone should raise an Error if called after
            # executing a query that cannot return rows
            cur.execute(
                f"{self.insert} into {self.table_prefix}booze"
                " values ('Victoria Bitter')")
            self.assertRaises(self.driver.Error, cur.fetchone)

            cur.execute(f'select name from {self.table_prefix}booze')
            r = cur.fetchone()
            self.assertEqual(
                len(r), 1,
                'cursor.fetchone should have retrieved a single row')
            self.assertEqual(
                r[0], 'Victoria Bitter',
                'cursor.fetchone retrieved incorrect data')
            self.assertIsNone(
                cur.fetchone(),
                'cursor.fetchone should return None if no more rows available')
            self.assertIn(cur.rowcount, (-1, 1))
        finally:
            con.close()

    def test_next(self):
        """Test extension for getting the next row."""
        con = self._connect()
        try:
            cur = con.cursor()
            if not hasattr(cur, 'next'):
                return

            # cursor.next should raise an Error if called before
            # executing a select-type query
            self.assertRaises(self.driver.Error, cur.next)

            # cursor.next should raise an Error if called after
            # executing a query that cannot return rows
            self.execute_ddl1(cur)
            self.assertRaises(self.driver.Error, cur.next)

            # cursor.next should return None if a query retrieves no rows
            cur.execute(f'select name from {self.table_prefix}booze')
            self.assertRaises(StopIteration, cur.next)
            self.assertIn(cur.rowcount, (-1, 0))

            # cursor.next should raise an Error if called after
            # executing a query that cannot return rows
            cur.execute(f"{self.insert} into {self.table_prefix}booze"
                        " values ('Victoria Bitter')")
            self.assertRaises(self.driver.Error, cur.next)

            cur.execute(f'select name from {self.table_prefix}booze')
            r = cur.next()
            self.assertEqual(
                len(r), 1,
                'cursor.fetchone should have retrieved a single row')
            self.assertEqual(
                r[0], 'Victoria Bitter',
                'cursor.next retrieved incorrect data')
            # cursor.next should raise StopIteration if no more rows available
            self.assertRaises(StopIteration, cur.next)
            self.assertIn(cur.rowcount, (-1, 1))
        finally:
            con.close()

    samples = (
        'Carlton Cold',
        'Carlton Draft',
        'Mountain Goat',
        'Redback',
        'Victoria Bitter',
        'XXXX'
    )

    def _populate(self):
        """Return a list of SQL commands to setup the DB for fetching tests."""
        populate = [
            f"{self.insert} into {self.table_prefix}booze values ('{s}')"
            for s in self.samples]
        return populate

    def test_fetchmany(self):
        con = self._connect()
        try:
            cur = con.cursor()

            # cursor.fetchmany should raise an Error if called without
            # issuing a query
            self.assertRaises(self.driver.Error, cur.fetchmany, 4)

            self.execute_ddl1(cur)
            for sql in self._populate():
                cur.execute(sql)

            cur.execute(f'select name from {self.table_prefix}booze')
            r = cur.fetchmany()
            self.assertEqual(
                len(r), 1,
                'cursor.fetchmany retrieved incorrect number of rows,'
                ' default of arraysize is one.')
            cur.arraysize = 10
            r = cur.fetchmany(3)  # Should get 3 rows
            self.assertEqual(
                len(r), 3,
                'cursor.fetchmany retrieved incorrect number of rows')
            r = cur.fetchmany(4)  # Should get 2 more
            self.assertEqual(
                len(r), 2,
                'cursor.fetchmany retrieved incorrect number of rows')
            r = cur.fetchmany(4)  # Should be an empty sequence
            self.assertEqual(
                len(r), 0,
                'cursor.fetchmany should return an empty sequence after'
                ' results are exhausted')
            self.assertIn(cur.rowcount, (-1, 6))

            # Same as above, using cursor.arraysize
            cur.arraysize = 4
            cur.execute(f'select name from {self.table_prefix}booze')
            r = cur.fetchmany()  # Should get 4 rows
            self.assertEqual(
                len(r), 4,
                'cursor.arraysize not being honoured by fetchmany')
            r = cur.fetchmany()  # Should get 2 more
            self.assertEqual(len(r), 2)
            r = cur.fetchmany()  # Should be an empty sequence
            self.assertEqual(len(r), 0)
            self.assertIn(cur.rowcount, (-1, 6))

            cur.arraysize = 6
            cur.execute(f'select name from {self.table_prefix}booze')
            rows = cur.fetchmany()  # Should get all rows
            self.assertIn(cur.rowcount, (-1, 6))
            self.assertEqual(len(rows), 6)
            self.assertEqual(len(rows), 6)
            rows = [r[0] for r in rows]
            rows.sort()

            # Make sure we get the right data back out
            for i in range(0, 6):
                self.assertEqual(
                    rows[i], self.samples[i],
                    'incorrect data retrieved by cursor.fetchmany')

            rows = cur.fetchmany()  # Should return an empty list
            self.assertEqual(
                len(rows), 0,
                'cursor.fetchmany should return an empty sequence if'
                ' called after the whole result set has been fetched')
            self.assertIn(cur.rowcount, (-1, 6))

            self.execute_ddl2(cur)
            cur.execute(f'select name from {self.table_prefix}barflys')
            r = cur.fetchmany()  # Should get empty sequence
            self.assertEqual(
                len(r), 0,
                'cursor.fetchmany should return an empty sequence if'
                ' query retrieved no rows')
            self.assertIn(cur.rowcount, (-1, 0))

        finally:
            con.close()

    def test_fetchall(self):
        con = self._connect()
        try:
            cur = con.cursor()
            # cursor.fetchall should raise an Error if called
            # without executing a query that may return rows (such
            # as a select)
            self.assertRaises(self.driver.Error, cur.fetchall)

            self.execute_ddl1(cur)
            for sql in self._populate():
                cur.execute(sql)

            # cursor.fetchall should raise an Error if called
            # after executing a a statement that cannot return rows
            self.assertRaises(self.driver.Error, cur.fetchall)

            cur.execute(f'select name from {self.table_prefix}booze')
            rows = cur.fetchall()
            self.assertIn(cur.rowcount, (-1, len(self.samples)))
            self.assertEqual(
                len(rows), len(self.samples),
                'cursor.fetchall did not retrieve all rows')
            rows = sorted(r[0] for r in rows)
            for i in range(0, len(self.samples)):
                self.assertEqual(
                    rows[i], self.samples[i],
                    'cursor.fetchall retrieved incorrect rows')
            rows = cur.fetchall()
            self.assertEqual(
                len(rows), 0,
                'cursor.fetchall should return an empty list if called'
                ' after the whole result set has been fetched')
            self.assertIn(cur.rowcount, (-1, len(self.samples)))

            self.execute_ddl2(cur)
            cur.execute(f'select name from {self.table_prefix}barflys')
            rows = cur.fetchall()
            self.assertIn(cur.rowcount, (-1, 0))
            self.assertEqual(
                len(rows), 0,
                'cursor.fetchall should return an empty list if'
                ' a select query returns no rows')

        finally:
            con.close()

    def test_mixedfetch(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.execute_ddl1(cur)
            for sql in self._populate():
                cur.execute(sql)

            cur.execute(f'select name from {self.table_prefix}booze')
            rows1 = cur.fetchone()
            rows23 = cur.fetchmany(2)
            rows4 = cur.fetchone()
            rows56 = cur.fetchall()
            self.assertIn(cur.rowcount, (-1, 6))
            self.assertEqual(
                len(rows23), 2,
                'fetchmany returned incorrect number of rows')
            self.assertEqual(
                len(rows56), 2,
                'fetchall returned incorrect number of rows')

            rows = [rows1[0]]
            rows.extend([rows23[0][0], rows23[1][0]])
            rows.append(rows4[0])
            rows.extend([rows56[0][0], rows56[1][0]])
            rows.sort()
            for i in range(0, len(self.samples)):
                self.assertEqual(
                    rows[i], self.samples[i],
                    'incorrect data retrieved or inserted')
        finally:
            con.close()

    def help_nextset_setup(self, cur):
        """Set up nextset test.

        Should create a procedure called deleteme that returns two result sets,
        first the number of rows in booze, then "name from booze".
        """
        raise NotImplementedError('Helper not implemented')
        # sql = """
        #    create procedure deleteme as
        #    begin
        #        select count(*) from booze
        #        select name from booze
        #    end
        # """
        # cur.execute(sql)

    def help_nextset_teardown(self, cur):
        """Clean up after nextset test.

        If cleaning up is needed after test_nextset.
        """
        raise NotImplementedError('Helper not implemented')
        # cur.execute("drop procedure deleteme")

    def test_nextset(self):
        """Test the nextset functionality."""
        raise NotImplementedError('Drivers need to override this test')
        # example test implementation only:
        # con = self._connect()
        # try:
        #     cur = con.cursor()
        #     if not hasattr(cur, 'nextset'):
        #         return
        #     try:
        #         self.executeDDL1(cur)
        #         for sql in self._populate():
        #             cur.execute(sql)
        #         self.help_nextset_setup(cur)
        #         cur.callproc('deleteme')
        #         number_of_rows = cur.fetchone()
        #         self.assertEqual(number_of_rows[0], len(self.samples))
        #         self.assertTrue(cur.nextset())
        #         names = cur.fetchall()
        #         self.assertEqual(len(names), len(self.samples))
        #         self.assertIsNone(
        #             cur.nextset(), 'No more return sets, should return None')
        #     finally:
        #         self.help_nextset_teardown(cur)
        # finally:
        #     con.close()

    def test_arraysize(self):
        # Not much here - rest of the tests for this are in test_fetchmany
        con = self._connect()
        try:
            cur = con.cursor()
            self.assertTrue(hasattr(cur, 'arraysize'),
                            'cursor.arraysize must be defined')
        finally:
            con.close()

    def test_setinputsizes(self):
        con = self._connect()
        try:
            cur = con.cursor()
            cur.setinputsizes((25,))
            self._paraminsert(cur)  # Make sure cursor still works
        finally:
            con.close()

    def test_setoutputsize_basic(self):
        # Basic test is to make sure setoutputsize doesn't blow up
        con = self._connect()
        try:
            cur = con.cursor()
            cur.setoutputsize(1000)
            cur.setoutputsize(2000, 0)
            self._paraminsert(cur)  # Make sure the cursor still works
        finally:
            con.close()

    def test_setoutputsize(self):
        # Real test for setoutputsize is driver dependant
        raise NotImplementedError('Driver needed to override this test')

    def test_none(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.execute_ddl2(cur)
            # inserting NULL to the second column, because some drivers might
            # need the first one to be primary key, which means it needs
            # to have a non-NULL value
            cur.execute(f"{self.insert} into {self.table_prefix}barflys"
                        " values ('a', NULL)")
            cur.execute(f'select drink from {self.table_prefix}barflys')
            r = cur.fetchall()
            self.assertEqual(len(r), 1)
            self.assertEqual(len(r[0]), 1)
            self.assertIsNone(r[0][0], 'NULL value not returned as None')
        finally:
            con.close()

    def test_date(self):
        d1 = self.driver.Date(2002, 12, 25)
        d2 = self.driver.DateFromTicks(
            time.mktime((2002, 12, 25, 0, 0, 0, 0, 0, 0)))
        # Can we assume this? API doesn't specify, but it seems implied
        self.assertEqual(str(d1), str(d2))

    def test_time(self):
        t1 = self.driver.Time(13, 45, 30)
        t2 = self.driver.TimeFromTicks(
            time.mktime((2001, 1, 1, 13, 45, 30, 0, 0, 0)))
        # Can we assume this? API doesn't specify, but it seems implied
        self.assertEqual(str(t1), str(t2))

    def test_timestamp(self):
        t1 = self.driver.Timestamp(2002, 12, 25, 13, 45, 30)
        t2 = self.driver.TimestampFromTicks(
            time.mktime((2002, 12, 25, 13, 45, 30, 0, 0, 0))
        )
        # Can we assume this? API doesn't specify, but it seems implied
        self.assertEqual(str(t1), str(t2))

    def test_binary_string(self):
        self.driver.Binary(b'Something')
        self.driver.Binary(b'')

    def test_string_type(self):
        self.assertTrue(hasattr(self.driver, 'STRING'),
                        'module.STRING must be defined')

    def test_binary_type(self):
        self.assertTrue(hasattr(self.driver, 'BINARY'),
                        'module.BINARY must be defined.')

    def test_number_type(self):
        self.assertTrue(hasattr(self.driver, 'NUMBER'),
                        'module.NUMBER must be defined.')

    def test_datetime_type(self):
        self.assertTrue(hasattr(self.driver, 'DATETIME'),
                        'module.DATETIME must be defined.')

    def test_rowid_type(self):
        self.assertTrue(hasattr(self.driver, 'ROWID'),
                        'module.ROWID must be defined.')
